#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "AVLTree.h"

// Returns the subtree height - This is equal to the maximum of the two children heights
int subtreeHeight(AVLTreeNode *node)
{
	int height = 0;

	if (node == NULL)
		return -1;
	else
		return node->height;
}

// Destroy a subtree
void deleteSubtree(AVLTreeNode *node)
{
	assert(node != NULL);
	// First destroy the children
	if (node->leftChild != NULL)
		deleteSubtree(node->leftChild);
	if (node->rightChild != NULL)
		deleteSubtree(node->rightChild);
	// Delete the node
	free(node);
}

// Create a new node
AVLTreeNode *createNode(Type value)
{
	AVLTreeNode *node = (AVLTreeNode *)malloc(sizeof(AVLTreeNode));

	if (node == NULL) {
		fprintf(stderr, "Memory Allocation Error! Exiting!\n");
		exit(1);
	}
	node->data = value;
	node->height = 0;
	node->leftChild = node->rightChild = NULL;
	node->father = NULL;

	return node;
}

// Print the subtree in infix order
void printNodeInfix(AVLTreeNode *node)
{
	assert(node != NULL);

	if (node->leftChild)
		printNodeInfix(node->leftChild);

	fprintf(stdout, TYPE_MODIFIER" ", node->data);

	if (node->rightChild)
		printNodeInfix(node->rightChild);
}

// Print the subtree in prefix order
void printNodePrefix(AVLTreeNode *node)
{
	assert(node != NULL);

	fprintf(stdout, TYPE_MODIFIER" ", node->data);

	if (node->leftChild)
		printNodePrefix(node->leftChild);
	if (node->rightChild)
		printNodePrefix(node->rightChild);
}

// Print the subtree in postfix order
void printNodePostfix(AVLTreeNode *node)
{
	assert(node != NULL);

	if (node->leftChild)
		printNodePostfix(node->leftChild);
	if (node->rightChild)
		printNodePostfix(node->rightChild);
	
	fprintf(stdout, TYPE_MODIFIER" ", node->data);
}

// Insert a node into the Binary Search Tree
AVLTreeNode *insertNode(AVLTreeNode *root, AVLTreeNode *node)
{
	assert(root != NULL);
	assert(node != NULL);

	if (node->data <= root->data) {

#ifdef _DEBUG
		fprintf(stdout, "(%2d)    left of: "TYPE_MODIFIER"\n", root->height, root->data);
#endif
		if (root->leftChild != NULL) {
			root->leftChild = insertNode(root->leftChild, node);
		}
		else {
			root->leftChild = node;
		}
		root->leftChild->father = root;
		// Rebalance the tree to ensure AVL structuring
		if ((subtreeHeight(root->leftChild) - subtreeHeight(root->rightChild) == 2)) {
			// The left subtree is taller than the right
			if (node->data <= root->leftChild->data) {
				// Single left rotation (LL)
				root = singleRotationLeft(root);
			}
			else {
				// Double rotation in the left child - The problem is left and then right (LR)
				root = doubleRotationLeftRight(root);
			}
		}
	}
	else {

#ifdef _DEBUG
		fprintf(stdout, "(%2d)   right of: "TYPE_MODIFIER"\n", root->height, root->data);
#endif

		if (root->rightChild != NULL) {
			root->rightChild = insertNode(root->rightChild, node);
		}
		else {
			root->rightChild = node;			
		}
		root->rightChild->father = root;
		// Rebalance the tree to ensure AVL structuring
		if ((subtreeHeight(root->leftChild) - subtreeHeight(root->rightChild) == -2)) {
			// The right subtree is taller than the left
			if (node->data > root->rightChild->data) {
				// Single right rotation - RR
				root = singleRotationRight(root);
			}
			else {
				// Double rotation in the right child - The problem is right and then left
				root = doubleRotationRightLeft(root);
			}
		}
	}
	// Update the subtree height
	root->height = 1 + MAX(subtreeHeight(root->leftChild), subtreeHeight(root->rightChild));

	return root;
}

/* The problem is LL, so do a rotation in the left child */
AVLTreeNode *singleRotationLeft(AVLTreeNode *root)
{
	AVLTreeNode *newRoot = NULL;

	// Assign the new root
	newRoot = root->leftChild;
	root->leftChild = newRoot->rightChild;
	newRoot->rightChild = root;
	newRoot->father = root->father;
	root->father = newRoot;

	// Fix the heights of the previous root node and the new root node
	root->height = MAX(subtreeHeight( root->leftChild ),
					   subtreeHeight( root->rightChild )) + 1;
	newRoot->height = MAX(subtreeHeight(newRoot->leftChild ),
						  subtreeHeight(newRoot->rightChild)) + 1;

#ifdef _DEBUG
	fprintf(stdout, "Rotation Left: \n\tMoved "TYPE_MODIFIER" up, and "TYPE_MODIFIER" down\n",
					newRoot->data, root->data);
#endif // _DEBUG

	// Return the new root
	return newRoot;
}

/* The problem is RR, so do a rotation in the right child */
AVLTreeNode *singleRotationRight(AVLTreeNode *root)
{
	AVLTreeNode *newRoot = NULL;

	// Assign the new root
	newRoot = root->rightChild;
	root->rightChild = newRoot->leftChild;
	newRoot->leftChild = root;
	newRoot->father = root->father;
	root->father = newRoot;

	// Fix the heights of the previous root node and the new root node
	root->height = MAX(subtreeHeight(root->leftChild ),
					   subtreeHeight(root->rightChild )) + 1;
	newRoot->height = MAX(subtreeHeight(newRoot->leftChild ),
						  subtreeHeight(newRoot->rightChild)) + 1;

#ifdef _DEBUG
	fprintf(stdout, "Rotation Right: \n\tMoved "TYPE_MODIFIER" up, and "TYPE_MODIFIER" down\n",
					newRoot->data, root->data);
#endif // _DEBUG

	return newRoot;
}


/**
 * The problem is LR, so do a rotation on the right subtree of the left child
 * and then a rotation on the left child.
 */	
AVLTreeNode *doubleRotationLeftRight(AVLTreeNode *root)
{
	AVLTreeNode *newRoot = NULL;

#ifdef _DEBUG
	fprintf(stdout, "LR rotation in left subtree for "TYPE_MODIFIER"\n", root->data);
#endif // _DEBUG

	// First rotate the right subtree of the roots left child
	root->leftChild = singleRotationRight(root->leftChild);
	// Then rotate the left child of the root
	newRoot = singleRotationLeft(root);

	return newRoot;
}

/**
 * The problem is RL, so do a rotation on the left subtree of the right child
 * and then a rotation on the right child.
 */
AVLTreeNode *doubleRotationRightLeft(AVLTreeNode *root)
{
	AVLTreeNode *newRoot = NULL;

#ifdef _DEBUG
	fprintf(stdout, "RL rotation in right subtree for node: "TYPE_MODIFIER"\n", root->data);
#endif // _DEBUG

	// First rotate the left subtree of the roots right child
	root->rightChild = singleRotationLeft(root->rightChild);
	// Then rotate the right child of the root
	newRoot = singleRotationRight(root);

	return newRoot;
}



// Find a node containing a data value
AVLTreeNode *findNode(AVLTreeNode *node, Type value)
{
	if (node == NULL)
		return NULL;

	if (node->data == value)
		return node;
	else if (value < node->data)
		return findNode(node->leftChild, value);
	else
		return findNode(node->rightChild, value);
}

// Delete the tree
void deleteTree(AVLTree *tree)
{
	assert(tree != NULL);

	if (tree->root != NULL)
		deleteSubtree(tree->root);
	free(tree);
}

// Create a new tree with no root
AVLTree *createTree(void)
{
	AVLTree *tree = NULL;

	tree = (AVLTree *)malloc(sizeof(AVLTree));
	if (tree == NULL) {
		fprintf(stderr, "Memory Allocation Error! Exiting!\n");
		exit(1);
	}

	tree->root = NULL;
	return tree;
}

// Insert the data into the node
void insertData(AVLTree *tree, Type value)
{
	AVLTreeNode *newNode = NULL;

	assert(tree != NULL);

	newNode = createNode(value);

#ifdef _DEBUG
	fprintf(stdout, "\nInserting Value \'"TYPE_MODIFIER"\'\n", value);
	fprintf(stdout, "------------------\n", value);
#endif

	// If there is no root, create a new one and assign the given value to it
	if (tree->root == NULL) {

#ifdef _DEBUG
		fprintf(stdout, "Root Node Created with value: "TYPE_MODIFIER"\n", value);
#endif
		tree->root = newNode;
	}
	else
		tree->root = insertNode(tree->root, newNode);
}

// Print the tree in multiple visit orders
void printTree(AVLTree *tree)
{
	assert(tree != NULL);

	fprintf(stdout, "Tree contents (infix): ");
	printNodeInfix(tree->root);
	fprintf(stdout,  "\n");

	fprintf(stdout, "Tree contents (prefix): ");
	printNodePrefix(tree->root);
	fprintf(stdout,  "\n");

	fprintf(stdout, "Tree contents (postfix): ");
	printNodePostfix(tree->root);
	fprintf(stdout,  "\n");
}

// Find a node that contains a data value
AVLTreeNode *findData(AVLTree *tree, Type value)
{
	AVLTreeNode *node = NULL;

	assert(tree != NULL);

	node = findNode(tree->root, value);

	return node;
}

// Checks if a data value exists in the tree
BOOL dataExists(AVLTree *tree, Type value)
{
	assert(tree != NULL);
	if (findData(tree, value) != NULL)
		return TRUE;
	else
		return FALSE;
}

BOOL deleteData(AVLTree *tree, Type value)
{
	AVLTreeNode *node = NULL;
	assert(tree != NULL);

	node = findData(tree, value);

	if (node == NULL)
		return FALSE;

	return TRUE;

}
